(* -------------------------------------------------------------*)
(*		EvalObj.ML: 					*)
(*		Derived Rules for Semantic Objects		*)
(* ------------------------------------------------------------ *)

structure eval_obj_lemmas =
struct
local open SML_Syntax eval_obj
in

(* -------------------------------------------------------------------- *)
(*	Derived Rules for Semantic Objects 				*)
(* -------------------------------------------------------------------- *)

(* type assertions for components of an environment  *)

fun type_tacfun prems = [cut_facts_tac prems 1 THEN
               REPEAT (eresolve_tac [SigmaE2] 1) THEN
               assume_tac 1];

val SE_type = rewrite_rule [StrEnv_def] 
  (prove_goalw thy [Env_def] 
    "<SE,VE,EE> : Env ==> SE : StrEnv" type_tacfun);

val VE_type = rewrite_rule [VarEnv_def] 
  (prove_goalw thy [Env_def] 
    "<SE,VE,EE> : Env ==> VE : VarEnv" type_tacfun);

val EE_type = rewrite_rule [ExConEnv_def] 
  (prove_goalw thy [Env_def] 
    "<SE,VE,EE> : Env ==> EE : ExConEnv" type_tacfun);


fun tacfun prems = [cut_facts_tac prems 1 THEN fast_tac ZF_cs 1];

val Var_type = prove_goal thy 
 "x : Var ==> <Var, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val Con_type = prove_goal thy 
 "x : Con ==> <Con, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val ExCon_type = prove_goal thy 
 "x : ExCon ==> <ExCon, x> : (SUM x:{Var,Con,ExCon}. x)" tacfun;

val  [SE0_type, VE0_type, EE0_type] =
    ([rewrite_rule [E0_def] E0_type] RL 
     [SE_type, VE_type, EE_type]) RL [mapD];
  
fun tacfun prems = 
	     [cut_facts_tac prems 1, 
              rtac apply_equality1 1,
              REPEAT (ares_tac [SE0_type, VE0_type, EE0_type] 1)];

val apply_SE0 = prove_goal thy "<arg, s> : SE0 ==> SE0`arg = s" tacfun ;

val apply_VE0 = prove_goal thy "<arg, s> : VE0 ==> VE0`arg = s" tacfun ;

val apply_EE0 = prove_goal thy "<arg, s> : EE0 ==> EE0`arg = s" tacfun ;

fun mk_mems def triv_thm =
let fun f th  = 
       (consI1 RS th) ::  f (consI2 RS th) handle _ => []
in map (rewrite_rule (map symmetric [def])) 
      (f (rewrite_rule [def] triv_thm))
end;

val VE0_mems = mk_mems VE0_def 
  (prove_goal thy "p : VE0 ==> p : VE0" 
            (fn prems => [cut_facts_tac prems 1, atac 1]));

val combine_equality = prove_goalw thy [env_combine_def]
 "[| SE @@ SE' = SE''; VE @@ VE' = VE''; EE @@ EE' = EE'' ; \
\    <SE'', VE'', EE''> = E |] ==>\
\    <SE, VE, EE> # <SE',VE', EE'> = E"
  (fn prems => [cut_facts_tac prems 1, asm_simp_tac ZF_ss 1]);

val E0_combine_equality = 
prove_goalw thy [env_combine_def, E0_def]
 "[|  SE0 @@ SE' = SE'' ;     VE0 @@ VE' = VE''; \
\    EE0 @@ EE' = EE'' ;    <SE'', VE'', EE''> = E |] ==> \
\    E0 # <SE', VE', EE'> = E" 
(fn prems => [cut_facts_tac prems 1, asm_simp_tac ZF_ss 1]);

(* -------------------------------------------------------------------- *)
(*	 	Resolution Tactics	  				*)
(* -------------------------------------------------------------------- *)

fun list_subst_vars [] [] = [] |
    list_subst_vars (th::thlist) (strlist::strlistlist) =
       (subst_vars th strlist) :: list_subst_vars thlist strlistlist;

fun get_intr_rules thy (name, strlistlist) =
  list_subst_vars (mk_intr_rule (get_axiom thy name)) strlistlist;

val Rec_rules 
 = [Rec_empty_rule, Rec_rule, Rec_function]; 

val combine_rules = 
 [combine_equality, E0_combine_equality];

val apply_intr_rules_data = 
  [ ("plus_rule", [[]]),
    ("minus_rule", [[]]),
    ("times_rule", [[]]),
    ("zero_rule", [["v"], ["v"]]),
    ("not_rule", [["v"], ["v"]]),
    ("head_rule", [["v"]]),
    ("tail_rule", [["L"]]),
    ("null_rule", [["v"], ["v"]]),
    ("cons_rule", [[]]),
    ("apply_con", [[]]),
    ("apply_excon", [[]]),
    ("apply_closure", [[],[]])];

val unflattened_apply_intr_rules = 
 map (get_intr_rules thy) apply_intr_rules_data;

val apply_intr_rules = 
 flat unflattened_apply_intr_rules;

val Val_rules =
   [Val_eq, Val_neq_pack, Val_neq_pack RS not_sym,
   BasVal_neq_closure, BasVal_neq_closure RS not_sym,
   ExVal_neq_closure, ExVal_neq_closure RS not_sym,
   BasVal_neq_ExVal, BasVal_neq_ExVal RS not_sym];

val init_rules = 
   [SE0_type, VE0_type, EE0_type,
    apply_SE0, apply_VE0, apply_EE0];
 
val Obj_rules = [closure_neq_empty, 
                 Var_type, Con_type, ExCon_type] @ 
                 init_rules @ Val_rules @ arith_rules @ 
                 combine_rules @ VE0_mems @ Rec_rules;

(* -------------------------------------------------------------------- *)
(*	 	 Solving Side-Conditions				*)
(* -------------------------------------------------------------------- *)

val distinct_rules = 
  [var_con_distinct, 
   sym RS var_con_distinct,
   var_excon_distinct,
   sym RS var_excon_distinct,
   con_excon_distinct, 
   sym RS con_excon_distinct];

val Eval_cs = Kernel_cs addSEs distinct_rules;

val apply_tac = resolve_tac apply_intr_rules;

val obj_step_tac =  kernel_step_tac Obj_rules Eval_cs;

(* obj_step_tac is equivalent to  kernel_step_tac Obj_rules *)

val obj_tac = DEPTH_SOLVE_1 o obj_step_tac;

end
end



