(* -------------------------------------------------------------*)
(*		Tools.ML: 					*)
(*		Miscellaneous Functions, 			*)
(*		including tools for forwards proof		*)
(* ------------------------------------------------------------ *)

fun first (x,y) = x and second (x,y) = y;

fun mk_pair_list [] [] = [] |
    mk_pair_list (a::b) (c::d) = (a,c)::mk_pair_list b d;

(* -------------------------------------------------------------------- *)
(*	Discriminators and Destructors for terms 			*)
(*	dest_term (Const (s,_) $ a1 $ ... $ an) = (s, [a1, ..., an])	*)
(* -------------------------------------------------------------------- *)

fun is_schvar (Var(_,_))  = true |  is_schvar _ = false;

local fun term_list (t $ t') = t' :: (term_list t) |
          term_list t = [t]
in
fun dest_term t = 
  let val (Const(s,_) :: rest) = rev (term_list t) in (s, rest) end 
end;

fun get_prop prem = let val ("Trueprop", [t]) =   
  dest_term (Logic.strip_assums_concl prem) in t end;


(* -------------------------------------------------------------------- *)
(*									*)
(*	ML functions for forwards proof					*)
(*									*)
(*	strip_exists:	  strips off all existential quantifiers	*)
(*									*)
(*	strip_conj:	  finds all assumptions of the form A & B	*)
(*			  and replaces them by [| A ; B |]		*)
(*									*)
(*	strip_disj:	  finds all assumptions of the form A | B	*)
(*			  and produces two new theorems (one which 	*)
(*			  assumes A, and one which assumes B).          *)
(*			  It always returns a list of theorems.		*)
(*									*)
(*      strip_conj_concl: breaks apart a conjunctive conclusion, and 	*)
(*			  returns a list of theorems 			*)
(*									*)
(*      strip_not_disj_concl: breaks apart a negated disjunction, and 	*)
(*			  returns a list of theorems 			*)
(*									*)
(*      strip_imp_concl:  strip implications from the conclusion	*)
(*									*)
(*      strip_not_concl:  simplifies a negated conclusion ~Q by 	*)
(*			  putting Q into the assumption list and	*)
(*			  making the conclusion anything.		*)
(*									*)
(* -------------------------------------------------------------------- *)

fun strip_exists_asm i th = strip_exists_asm i (exI RSN (i, th)) 
                        handle THM _ => th;

fun strip_exists th = let val i = ref 1;
                          val x = ref th 
                    in 
                       while !i <= length(prems_of (th)) do
                      (x := strip_exists_asm (!i) (!x);
                       i := !i+1); !x
                    end;

fun strip_conj_asm i th = conjI RSN (i, th) 
                        handle THM _ => th;

fun strip_conj th = let val i = ref 1;
                        val x = ref th 
                    in 
                       while !i <= length(prems_of (!x)) do
                      (x := strip_conj_asm (!i) (!x);
                       i := !i+1); !x
                    end;

(* this is an awful hack *)


fun strip_disj_asm i th = 
 let  fun resolution (state, i, rules) =
         biresolution false (map (fn rl => (false,rl)) rules) i state;
       val [th1, th2] = 
           Sequence.list_of_s (resolution (th,i,[disjI1,disjI2]))
 in (th1 ::  strip_disj_asm i th2)
 end
  handle _ => [th];
                   
fun strip_disj th = let val i = ref 1;
                        val x = ref [] 
                    in 
                       while !i <= length(prems_of (th)) do
                      (x := (!x) @ (strip_disj_asm (!i) th);
                       i := !i+1); (if null (!x) then [th] else (!x))
                    end;


fun strip_equal_asm i th = 
  let val x = (refl RSN (i,th))
  in (strip_equal_asm i x)
  end
 handle THM _ => th; 

(* not needed at the moment?  *)

fun strip_equal th = let val i = ref 1;
                         val x = ref th 
                    in 
                       while !i <= length(prems_of (th)) do
                      (x := strip_equal_asm (!i) (!x);
                       i := !i+1); !x
                    end;
                
val strip = (map ((* strip_equal o *)strip_conj o strip_exists)) o
                    strip_disj o strip_exists;

fun strip_conj_concl th = 
   (th RS conjunct1)::(strip_conj_concl (th RS conjunct2))
                          handle THM _ => [th];
                          
val cs = ZF_cs addIs [equalityI] addEs [equalityE];

val tacfun = fn prems => [cut_facts_tac prems 1, fast_tac cs 1];

val not_disj1 = prove_goal ZF.thy 
 "~ (a | b)  ==> ~a" tacfun;

val not_disj2 = prove_goal ZF.thy 
 "~ (a | b)  ==> ~b" tacfun;

fun strip_not_disj_concl th = 
   (th RS not_disj1)::(strip_not_disj_concl (th RS not_disj2))
                          handle THM _ => [th];

fun strip_imp_concl th = strip_imp_concl (th RS mp) handle THM _ => th;

fun strip_not_concl th =  (th RS notE) handle THM _ => th;


fun gen_sym_asm i th = [sym RSN (i,th), th] handle THM _ => [];

fun gen_sym th = let val i = ref 1;
                     val x = ref [] 
                 in 
                     while !i <= length(prems_of (th)) do
                    (x := (!x) @ (gen_sym_asm (!i) th);
                     i := !i+1); !x
                 end;
 
(* -------------------------------------------------------------------- *)
(*		Introduction Rules					*)
(* -------------------------------------------------------------------- *)

fun mk_intr_rule def = 
     strip ((rewrite_rule [iff_def] def) RS conjunct1 RS mp);

fun mk_intr_rules thmlist = flat (map mk_intr_rule thmlist);

fun subst_var th s =
let val prems = prems_of th;
    fun  f (p::ps, i) s =
        let val (_ $ (Const("op =", _) $ _ $ Var ((s', _),_))) = p
        in
           if s = s' then i else f (ps, i+1) s
        end
        handle _ => f (ps, i+1) s
in
    refl RSN (f (prems, 1) s, th)
end
handle _ => th;

fun subst_vars th [] = th |
    subst_vars th (a::alist)  = subst_vars (subst_var th a) alist;

(* -------------------------------------------------------------------- *)
(*		Elimination Rules					*)
(* -------------------------------------------------------------------- *)

val ex2E = prove_goal ZF.thy
 "[| EX x y. P(x,y); !! x y. P(x,y) ==> R |] ==> R"
 (fn prems =>
  [ (cut_facts_tac prems 1),
    (REPEAT (eresolve_tac [exE] 1 ORELSE ares_tac prems 1)) ]);

fun mk_elim_rule def = 
   let val th = ((rewrite_rule [iff_def] def) RS conjunct2 RS mp)
   in
      th RS exE
   handle _ => th RS ex2E 
   handle _ => th
end;

fun mk_elim_rules thmlist = map mk_elim_rule thmlist;

